from .wc_object import WcObj
from .common import WcContent
from requests.exceptions import HTTPError

class WcWorkable(WcObj):
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)
        
    def checkin(self, checkin_note=None):
        action = f'PTC.{self.domain}.CheckIn'
        body = {}
        if checkin_note:
            body['CheckInNote'] = checkin_note
        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=action,
            body=body
        )
        self.id = response.json()['ID']
        return response

    def checkout(self):
        action = f'PTC.{self.domain}.CheckOut'
        body = {}
        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=action,
            body=body
        )
        self.id = response.json()['ID']
        return response

    def move(self, context_id, folder_id):
        #Not OOTB Windchill endpoint
        move_domain = 'MoveObject'
        move_endpoint = 'moveObjects'
        move_data = {'MoveInfo': {
            'ObjectID': self.id,
            'ContextID': context_id,
            'FolderID': folder_id
            } }
        response = self.session.call_api(
            request_type='post',
            domain=move_domain,
            endpoint=move_endpoint,
            body=move_data
        )
        return response

    def revise(self, new_version):
        action = f'PTC.{self.domain}.Revise'
        body = {'VersionId': new_version}
        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=action,
            body=body
        )
        self.id = response.json()['ID']
        return response

    def set_state(self, new_state):
        action = f'PTC.{self.domain}.SetState'
        body = {'State': {'Value': new_state}}
        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=action,
            body=body
        )
        self.id = response.json()['ID']
        return response
        
class WcMfrPart(WcWorkable):
    domain = 'ProdMgmt'
    endpoint = 'ManufacturerParts'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

class WcVendPart(WcWorkable):
    domain = 'ProdMgmt'
    endpoint = 'VendorParts'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)
        
class WcPart(WcWorkable):
    domain = 'ProdMgmt'
    endpoint = 'Parts'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

    def get_desc_by_docs(self, select_str=''): #TODO: fix select_str. Currently returns an error when populated. 
        navigation = 'DescribedBy'
        expand_str = 'DescribedBy'
        part_desc_links = self.get_related(navigation, WcObj, select_str, expand_str=expand_str)
        return part_desc_links
    
    def add_desc_by_doc(self, doc):
        #Not OOTB Windchill endpoint
        action = 'DescribedBy'
        descby_body = {"DescribedBy@odata.bind" : f"Documents(\'{doc.id}\')"}

        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=action,
            body=descby_body
        )
        return response

    def del_desc_by_doc(self, part_desc_link):
        #Not OOTB Windchill endpoint
        action = f"DescribedBy('{part_desc_link}')"

        response = self.session.call_api(
            request_type='delete',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=action
        )
        return response

    def get_assoc_epms(self, select_str=''):
        navigation = 'PartDocAssociations'
        expand_str = 'RelatedCADDoc'
        part_doc_assocs = self.get_related(navigation, WcObj, select_str, expand_str=expand_str)
        return part_doc_assocs
    
    def add_assoc_epm(self, epm, assoc_type):
        add_assoc_endpoint = 'CreateAssociations'
        add_assoc_body = {"PartDocAssociations": [{"AssociationType":{"Value": f"{assoc_type}"}, "RelatedPart@odata.bind": f"Parts('{self.id}')", "RelatedCADDoc@odata.bind": f"CADDocuments('{epm.id}')" }]}

        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=add_assoc_endpoint,
            body=add_assoc_body
        )
        return response
    
    def del_assoc_epm(self, epm, assoc_type):
        del_assoc_endpoint = 'DeleteAssociations'
        del_assoc_body = f'{{"PartDocAssociations": [{{"AssociationType":{{"Value": "{assoc_type}"}}, "RelatedPart@odata.bind": "Parts(\'{self.id}\')", "RelatedCADDoc@odata.bind": "CADDocuments(\'{epm.id}\')" }}]}}'

        response = self.session.call_api(
            request_type='post',
            domain=self.domain,
            endpoint=del_assoc_endpoint,
            body=del_assoc_body
        )
        return response
    
    def get_reps(self):
        navigation = 'Representations'
        objs = self.get_related(navigation, WcObj)
        return objs
        
class WcDoc(WcWorkable):
    domain = 'DocMgmt'
    endpoint = 'Documents'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)
    
    def get_prim_content(self, select_str=''):
        navigation = 'PrimaryContent'
        obj = self.get_related_single_obj(navigation, WcContent, select_str)
        return obj
    
    def get_attachments(self, select_str=''):
        navigation = 'Attachments'
        objs = self.get_related(navigation, WcContent, select_str)
        return objs
    
    def add_prim_content(self, filename, file_bytes):
        try:
            self.del_prim_content()
        except HTTPError as e:
            pass
        response = WcContent.upload_file(
            wc_session=self.session,
            domain=self.domain,
            endpoint=self.endpoint,
            obj_id=self.id,
            filename=filename,
            file_bytes=file_bytes,
            primary_content=True
        )
        return response
        
    def del_prim_content(self):
        navigation = 'PrimaryContent'
        response = self.session.call_api(
            request_type='delete',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=navigation
        )
        return response
    
    def add_attachment(self, filename, file_bytes):
        response = WcContent.upload_file(
            wc_session=self.session,
            domain=self.domain,
            endpoint=self.endpoint,
            obj_id=self.id,
            filename=filename,
            file_bytes=file_bytes,
            primary_content=False
        )
        return response
        
class WcEpm(WcWorkable):
    domain = 'CADDocumentMgmt'
    endpoint = 'CADDocuments'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

    def get_reps(self):
        navigation = 'Representations'
        objs = self.get_related(navigation, WcRep)
        return objs


from .wc_object import WcObj

class WcContent(WcObj):
    domain = 'PTC' #There may not be an exposed domain/endpoint to retrieve this object directly
    endpoint = 'ContentItem' #There may not be an exposed domain/endpoint to retrieve this object directly
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)
	
    @staticmethod
    def upload_file(wc_session, domain, endpoint, obj_id, filename, file_bytes, primary_content=False):
        stage1_action = f'PTC.{domain}.UploadStage1Action'
        stage1_body = {'NoOfFiles': 1}
        response = wc_session.call_api(
            request_type='post',
            domain=domain,
            endpoint=endpoint,
            arguments=obj_id,
            action=stage1_action,
            body=stage1_body
        )
        stage2_url = response.json()['value'][0]['ReplicaUrl']
        stream_id = response.json()['value'][0]['StreamIds'][0]
        wc_file_name = response.json()['value'][0]['FileNames'][0]
        
        stage2_addl_headers = {'Content-Type': 'multipart/form-data; boundary=---------------------------boundary'}
        stage2_body_part1 = f'''-----------------------------boundary
Content-Disposition: form-data; name="Master_URL"

{wc_session.base_url}/WindchillGW
-----------------------------boundary
Content-Disposition: form-data; name="CacheDescriptor_array"

{stream_id}:{wc_file_name}:{stream_id}
-----------------------------boundary
Content-Disposition: form-data; name="{wc_file_name}"; filename="{filename}"

'''#file_bytes will go here
        stage2_body_part2 = '''
-----------------------------boundary
'''
        #Note: String above ^^^ is very finicky - even spaces will break it. Also, note that there are 2 more dashes in header than in ---...boundary in body. 
        stage2_body = stage2_body_part1.encode() + file_bytes + stage2_body_part2.encode()
        response = wc_session.call_api(
            request_type='post',
            raw_url = stage2_url,
            data=stage2_body, #TODO: See if we can pass this in a json object. then we can remove the data input from the call_api signature. 
            addl_headers=stage2_addl_headers
        )
        file_size = response.json()['contentInfos'][0]['fileSize']
        encoded_info = response.json()['contentInfos'][0]['encodedInfo']
        
        stage3_action = f'PTC.{domain}.UploadStage3Action'
        stage3_body = {
            "ContentInfo" : [
                {	
                    "StreamId" : stream_id,
                    "EncodedInfo" : encoded_info,
                    "FileName" : filename,
                    "PrimaryContent" : primary_content,
                    "MimeType" : "null",
                    "FileSize" : file_size
                }
            ]
        }
        response = wc_session.call_api(
            request_type='post',
            domain=domain,
            endpoint=endpoint,
            arguments=obj_id,
            action=stage3_action,
            body=stage3_body
        )
        return response
        
    def download_file(self):
        file_url = self.attrs['Content']['URL']
        response = self.session.call_api(
            request_type='get',
            raw_url=file_url
        )
        return response.content
    
class WcRep(WcObj):
    domain = 'PTC' #There may not be an exposed domain/endpoint to retrieve this object directly
    endpoint = 'Representation' #There may not be an exposed domain/endpoint to retrieve this object directly
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)


from .wc_object import WcObj

class WcPrincipal(WcObj):
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

class WcGroup(WcPrincipal):
    domain = 'PrincipalMgmt'
    endpoint = 'Groups'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)
    
    def get_users(self, select_str=''):
        users = self.get_related('Users', WcUser, select_str)
        return users
        
class WcUser(WcPrincipal):
    domain = 'PrincipalMgmt'
    endpoint = 'Users'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)



from .wc_object import WcObj
from .workable import WcWorkable

class WcChange(WcObj):
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

    def get_aff_objs(self):
        navigation = 'AffectedObjects'
        objs = self.get_related(navigation, WcWorkable)
        return objs


class WcPr(WcChange):
    domain = 'ChangeMgmt'
    endpoint = 'ProblemReports'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

class WcCr(WcChange):
    domain = 'ChangeMgmt'
    endpoint = 'ChangeRequests'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)

class WcCn(WcChange):
    domain = 'ChangeMgmt'
    endpoint = 'ChangeNotices'
    def __init__(self, session, obj_id=None, json=None):
        super().__init__(session, obj_id, json)
    
    def get_res_objs(self):
        navigation = 'ResultingObjects'
        objs = self.get_related(navigation, WcWorkable)
        return objs



import requests
import webbrowser
import time
from urllib.parse import urlparse, parse_qs

class WcSession:
    def __init__(self, wc_base_url, auth_type, auth_params):
        self.auth_type = auth_type
        
        if self.auth_type == 'basic':
            self.api_base_url = wc_base_url + '/servlet/odata'
            self.username = auth_params[0]
            self.password = auth_params[1]
        elif self.auth_type == 'oauth':
            self.api_base_url = wc_base_url + '/oauth/servlet/odata'
            self.auth_url = auth_params[0]
            self.token_url = auth_params[1]
            self.client_id = auth_params[2]
            self.client_secret = auth_params[3]
            self.redirect_uri = auth_params[4]
            self.scope = auth_params[5]
            self.access_token = None
            self.refresh_token = None
            self.token_expiry = 0
        else:
            raise ValueError("Invalud auth type. Should be 'basic' or 'oauth'.")
        
        self.headers = {}
        self.session = requests.Session()
        self.CSRF_DOMAIN = 'PTC'
        self.CSRF_ENDPOINT = 'GetCSRFToken'

    @classmethod
    def basic(cls, wc_base_url, username, password):
        return cls(wc_base_url, auth_type='basic', auth_params=(username, password))
    
    @classmethod
    def oauth(cls, wc_base_url, auth_url, token_url, client_id, client_secret, redirect_uri, scope):
        return cls(wc_base_url, auth_type='oauth', auth_params=(auth_url, token_url, client_id, client_secret, redirect_uri, scope))
        
    def authenticate(self):
        if self.auth_type == 'basic':
            self.session.auth = (self.username, self.password)
        elif self.auth_type == 'oauth':
            auth_request_url = f'{self.auth_url}?client_id={self.client_id}&response_type=code&redirect_uri={self.redirect_uri}&scope={self.scope}'
            webbrowser.open(auth_request_url)
            print("Please authorize the application and provide the URL that you are redirected to.")
            redirect_url = input("Enter the redirect URL: ")
            qs = parse_qs(urlparse(redirect_url).query)
            auth_code = qs['code'][0]
            response = self.session.post(self.token_url, data={
                'grant_type': 'authorization_code',
                'code': auth_code,
                'redirect_uri': self.redirect_uri,
                'client_id': self.client_id,
                'client_secret': self.client_secret
            })
            if response.status_code != 200:
                raise self.session.exceptions.HTTPError(response.text)
            tokens = response.json()
            self.access_token = tokens['access_token']
            self.refresh_token = tokens['refresh_token']
            self.token_expiry = time.time() + tokens['expires_in']
            self.headers = {'Authorization': f'Bearer {self.access_token}'}
        else:
            raise ValueError("Invalud auth type. Should be 'basic' or 'oauth'.")

        response = self.call_api('get', domain=self.CSRF_DOMAIN, endpoint=self.CSRF_ENDPOINT)
        if response.status_code != 200:
            raise requests.exceptions.HTTPError(response.text)
        csrf = response.json()
        self.headers[csrf['NonceKey']] = csrf['NonceValue']

    def refresh(self):
        if self.auth_type == 'oauth' and time.time() > self.token_expiry:
            response = self.session.post(self.token_url, data={
                'grant_type': 'refresh_token',
                'refresh_token': self.refresh_token,
                'client_id': self.client_id,
                'client_secret': self.client_secret
            })
            tokens = response.json()
            self.access_token = tokens['access_token']
            self.token_expiry = time.time() + tokens['expires_in']
            self.headers['Authorization'] = f'Bearer {self.access_token}'

    def call_api(self, request_type, domain=None, endpoint=None, arguments=None, action=None, params=None, body=None, data=None, raw_url=None, addl_headers=None):
        #TODO: See if attachment upload can use json instead of data so we can remove the data from this signature
        self.refresh()
        if raw_url:
            url = raw_url
            params=None
        else:
            url = f"{self.api_base_url}/{domain}/{endpoint}"
            if arguments:
                url = url + f"(\'{arguments}\')"
            if action:
                url = url + f"/{action}"
        if addl_headers:
            headers = self.headers | addl_headers
        else:
            headers = self.headers

        if request_type.lower() == "get":
            #print(self.session.get(url, headers=self.headers, params=params).url)
            #print(params)
            #print(headers)
            response = self.session.get(url, headers=headers, params=params)
        elif request_type.lower() == "post":
            if data:
                response = self.session.post(url, headers=headers, params=params, data=data)
            else:
                response = self.session.post(url, headers=headers, params=params, json=body)
        elif request_type.lower() == "patch":
            response = self.session.patch(url, headers=headers, params=params, json=body)
        elif request_type.lower() == "delete":
            response = self.session.delete(url, headers=headers, params=params)
        else:
            raise ValueError(f"Unsupported request type {request_type}")
        
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            e.args = (f'{e.args[0]}\nError Text:\n{response.text}',)
            raise
        
        return response

    def get_objects(self, cls, filter_str="", select_str="", orderby_str="", expand_str="", params={}, page_size=None):
        #optional params: 'ptc.search.allowFilterOnTime': 'true', 'ptc.search.latestversion': 'true', 
        if filter_str:
            params["$filter"] = filter_str
        if select_str:
            params["$select"] = select_str
        if orderby_str:
            params['$orderby'] = orderby_str
        if expand_str:
            params['$expand'] = expand_str

        if page_size:
            addl_headers = {'Prefer': f'odata.maxpagesize={page_size}'}
        else:
            addl_headers = None
        
        domain = cls.domain
        endpoint = cls.endpoint

        response = self.call_api(
            request_type="get",
            domain=domain,
            endpoint=endpoint,
            params=params,
            addl_headers=addl_headers
        )
        data = response.json()
        
        for item in data.get("value", []):
            obj = cls(self, json=item)
            yield obj

        next_link = data.get("@odata.nextLink")
        while next_link:
            response = self.call_api(
                request_type="get",
                raw_url=next_link,
                addl_headers=addl_headers
            )
            data = response.json()
            
            for item in data.get("value", []):
                obj = cls(self, json=item)
                yield obj
            next_link = data.get("@odata.nextLink")

    def create_object(self, cls, attrs):
        response = self.call_api(
            request_type="post",
            domain=cls.domain,
            endpoint=cls.endpoint,
            body=attrs
        )
        obj = cls(self, json=response.json())
        return obj


class WcObj:

    def __init__(self, session, obj_id=None, json=None):
        self.session = session
        if json:
            self.attrs = json
            self.id = self.attrs['ID']

        if obj_id:
            response = self.session.call_api(
                request_type="get",
                domain=self.domain,
                endpoint=self.endpoint,
                arguments=obj_id
            )
            self.attrs = response.json()
            self.id = self.attrs['ID']

    def update(self, new_attrs):
        response = self.session.call_api(
            request_type='patch',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            body=new_attrs
        )
        return response
        
    def delete(self):
        response = self.session.call_api(
            request_type='delete',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
        )
        return response
        
    def get_related(self, navigation, cls, select_str='', expand_str=''):
        params = {}
        if select_str:
            params["$select"] = select_str
        if expand_str:
            params['$expand'] = expand_str
        
        response = self.session.call_api(
            request_type='get',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=navigation,
            params=params
        )
        data = response.json()
        
        for item in data.get("value", []):
            obj = cls(self.session, json=item)
            yield obj

        next_link = data.get("@odata.nextLink")
        while next_link:
            response = self.session.call_api(
                request_type='get',
                raw_url=next_link,
            )
            data = response.json()
            
            for item in data.get("value", []):
                obj = cls(self.session, json=item)
                yield obj
            next_link = data.get("@odata.nextLink")

    def get_related_single_obj(self, navigation, cls, select_str=''):
        if select_str:
            params = {'$select': select_str}
        else:
            params = None
        
        response = self.session.call_api(
            request_type='get',
            domain=self.domain,
            endpoint=self.endpoint,
            arguments=self.id,
            action=navigation,
            params=params
        )
        item = response.json()
        obj = cls(self.session, json=item)
        return obj
        
    def __str__(self):
        return f"{type(self).__name__}\nid: {self.id}\nNumber: {self.attrs['Number']}\nName: {self.attrs['Name']}"
